// DEBUG ---------------------------------------------------------------------

// INCLUDES ------------------------------------------------------------------
#include "relay2cl.h"
#include "spiswitch.h"

// MACROS --------------------------------------------------------------------

// DEFINES -------------------------------------------------------------------
//#define OPTION_HELP

#define MSG_OK "ok"
#define MSG_ERROR "error"
#define MSG_DELIMITER " \r\n"
#define NCMD 11

#define TRAP_LEFT 'l'
#define TRAP_RIGHT 'r'
#define	TRAP_LEFT_MAX 20
#define TRAP_RIGHT_MAX 20
#define TRAP_ID_A 'a'
#define TRAP_ID_B 'b'
#define TRAP_N 40
#define TRAP_CMD_ALL "all"

// TYPES ---------------------------------------------------------------------
enum ABusStateCode {ABUS_to_ZOUT, ABUS_to_BIAS};
enum AComStateCode {ACOM_to_ZIN, ACOM_to_GND};
enum BiasStateCode {BIAS_to_INT, BIAS_to_EXT};
enum ModeCode {MODE_EXC, MODE_Z, MODE_DEP};

// GLOBALS -------------------------------------------------------------------
const char Commands[NCMD][8] = {
	"help",		// 0
	"acom",		// 1
	"abus",		// 2
	"bias",		// 3
	"on",		// 4
	"off",		// 5
	"only",		// 6
	"status",	// 7
	"mode",		// 8
	"short",	// 9
	"open"};	// 10

	
enum CommandCode {
	CMD_HELP,		// 0
	CMD_ACOM,		// 1
	CMD_ABUS,		// 2
	CMD_BIAS,		// 3
	CMD_ON,			// 4
	CMD_OFF,		// 5
	CMD_ONLY,		// 6
	CMD_STATUS,		// 7
	CMD_MODE,		// 8
	CMD_SHORT,		// 9
	CMD_OPEN,		// 10
	CMD_INVALID		// 11
};

const char ABusStates[2][5] = {"zout", "bias"};
ABusStateCode ABusState = ABUS_to_ZOUT;

const char AComStates[2][4] = {"zin", "gnd"};
AComStateCode AComState = ACOM_to_ZIN;

const char BiasStates[2][4] = {"int", "ext"};
BiasStateCode BiasState = BIAS_to_INT;

const char Modes[3][4] = {"exc", "z", "dep"};

bool TrapState[TRAP_N] = {false};

char CMDBuffer[128];
uint8_t BufIndex = 0;

TwoCoilRelay ACOM(11, 10), ABUS(7, 6), BIAS(9, 8);
SPISwitch SwitchChain(SPISettings (15000000, MSBFIRST, SPI_MODE2), 4, 5, 5);

void setup()
{
	BIAS.Select(BIAS_to_INT);
	ACOM.Select(AComState);
	ABUS.Select(ABusState);
	
	Serial.begin(115200);
	Serial.println("ok");
	
	SwitchChain.Reset();
}

void loop()
{
	CommandCode cmd;
	while (Serial.available())
	{
		CMDBuffer[BufIndex] = Serial.read();
		++BufIndex;
	}

	if (BufIndex>1 && CMDBuffer[BufIndex-1] == '\n')
	{
		CMDBuffer[BufIndex-1] = '\0';
		BufIndex = 0;
		
		cmd = getCommand(CMDBuffer);
		switch (cmd)
		{
			#ifdef OPTION_HELP
			case CMD_HELP:
				handleHelp();
				break;
			#endif
			
			case CMD_ABUS:
				handleABus();
				break;
				
			case CMD_ACOM:
				handleACom();
				break;
				
			case CMD_BIAS:
				handleBias();
				break;
				
			case CMD_ON:
				handleOnOff(true);
				break;
			
			case CMD_OFF:
				handleOnOff(false);
				break;
				
			case CMD_ONLY:
				handleOnly(true);
				break;
				
			case CMD_STATUS:
				handleStatus();
				break;
				
			case CMD_MODE:
				handleMode();
				break;
				
			case CMD_SHORT:
				handleShort();
				break;
				
			case CMD_OPEN:
				handleOpen();
				break;
				
			default:
				Serial.println(MSG_ERROR);
				break;
		}
	}
}

int8_t getTrap(char *name)
{
	int8_t max;
	int8_t offset;
	char side, id;
	int8_t num;
	
	if (sscanf(name, "%c%d%c", &side, &num, &id) != 3) return -1;
	
	if (side == TRAP_LEFT)
	{
		max = TRAP_LEFT_MAX;
		offset = 0;
	}
	else if (side == TRAP_RIGHT)
	{
		max = TRAP_RIGHT_MAX;
		offset = TRAP_LEFT_MAX;
	}
	else
	{
		return -1;
	}
	
	if (num < 1 || num > max)
	{
		return -1;
	}
	
	num += offset -1;
	
	if (id == TRAP_ID_A)
	{
		Serial.println(MSG_ERROR);
	}
	else if (id == TRAP_ID_B)
	{
		offset = 1;
	}
	else
	{
		return -1;
	}
	
	return num;
}

void getTrapName(char *str, int8_t trap)
{
	char side, id;
	
	if (trap%2 == 0)
	{
		id = TRAP_ID_A;
	}
	else
	{
		id = TRAP_ID_B;
		trap -= 1;
	}
	
	if (trap < 2*TRAP_LEFT_MAX)
	{
		side = TRAP_LEFT;
		trap = 1 + trap/2;
	}
	else
	{
		side = TRAP_RIGHT;
		trap = 1 + trap/2 - TRAP_LEFT_MAX;
	}
	
	sprintf(str, "%c%02d%c", side, trap, id);
}

CommandCode getCommand(char *cmdString)
{
	uint8_t cmd;
	char *cmdWord = strtok(cmdString, MSG_DELIMITER);
	
	for (cmd=0; cmd<NCMD; ++cmd)
	{
		if (strcmp(cmdWord, Commands[cmd]) == 0)
		{
			break;
		}
	}
	
	return cmd;
}

void printSeparator()
{
	Serial.println("----------------------------------------------------------------------");
}

#ifdef OPTION_HELP
void handleHelp()
{
	char *tok;
	
	tok = strtok(NULL, MSG_DELIMITER);
	
	if (tok == NULL)
	{
		printSeparator();
		Serial.println("Valid commands:");
		Serial.println("");
		for (uint8_t i=0; i<NCMD; ++i)
		{
			Serial.println(Commands[i]);
		}
		Serial.println("");
		Serial.println("type \"help command\" for information on a specific command.");
		printSeparator();
	}
	else
	{
		uint8_t cmd;
		for (cmd=0; cmd<NCMD; ++cmd)
		{
			if (strcmp(tok, Commands[cmd]) == 0)
			{
				break;
			}
		}
		
		switch (cmd)
		{
			case CMD_HELP:
				printSeparator();
				Serial.println("help [command]");
				Serial.println("Prints the list of available commands.");
				Serial.println("If the command argument is specified, prints specific help for that command.");
				printSeparator();
				break;
				
			case CMD_ABUS:
				printSeparator();
				Serial.println("abus [zout|bias]");
				Serial.println("Returns the current ABUS connection state (zout or bias).");
				Serial.println("Sets the ABUS connection state accordingly if specified.");
				printSeparator();
				break;
			
			case CMD_ACOM:
				printSeparator();
				Serial.println("acom [zin|gnd]");
				Serial.println("Returns the current ACOM connection state (zin or gnd).");
				Serial.println("Sets the ACOM connection state accordingly if specified.");
				printSeparator();
				break;
				
			case CMD_BIAS:
				printSeparator();
				Serial.println("bias [int|ext]");
				Serial.println("Returns the current BIAS connection state (int or ext).");
				Serial.println("Sets the BIAS connection state accordingly if specified.");
				printSeparator();
				break;
				
			case CMD_ON:
				printSeparator();
				Serial.println("on [site1] [site2] ... [siteN]");
				Serial.println("Lists the currently ON sites. Or turns ON the specified sites.");
				Serial.println("Use \"on all\" to turn ON all the sites at once.");
				Serial.println("Site format: (l|r)(1-20)(a|b)\tExample: r14b");
				printSeparator();
				break;
				
			case CMD_OFF:
				printSeparator();
				Serial.println("off [site1] [site2] ... [siteN]");
				Serial.println("Lists the currently OFF sites. Or turns OFF the specified sites.");
				Serial.println("Use \"off all\" to turn OFF all the sites at once.");
				Serial.println("Site format: (l|r)(1-20)(a|b)\tExample: r14b");
				printSeparator();
				break;
				
			case CMD_ONLY:
				printSeparator();
				Serial.println("only site1 [site2] ... [siteN]");
				Serial.println("Turns ON the specified sites and OFF all the others.");
				Serial.println("Site format: (l|r)(1-20)(a|b)\tExample: r14b");
				printSeparator();
				break;
			
			case CMD_STATUS:
				printSeparator();
				Serial.println("status");
				Serial.println("Returns the complete connection status of the interface.");
				printSeparator();
				break;
				
			case CMD_MODE:
				printSeparator();
				Serial.println("mode (exc|z|dep)");
				Serial.println("Sets ACOM/ABUS/BIAS combinations for specific usages.");
				Serial.println("     +----+----+----+");
				Serial.println("     |ACOM|ABUS|BIAS|");
				Serial.println("+----+----+----+----+");
				Serial.println("|exc |gnd |bias|int | Site excitation from internal bias source.");
				Serial.println("+----+----+----+----+");
				Serial.println("|z   |zin |zout|int | Site connection to Z in/out connectors.");
				Serial.println("+----+----+----+----+");
				Serial.println("|dep |gnd |bias|ext | Site connection to external bias source.");
				Serial.println("+----+----+----+----+");
				printSeparator();
				break;
			
			default:
				Serial.print("Unknown command \"");
				Serial.print(tok);
				Serial.println("\".");
				break;
		}
	}

}
#endif

void handleABus()
{
	char *tok;
	
	tok = strtok(NULL, MSG_DELIMITER);
	
	if (tok == NULL)
	{
		Serial.println(ABusStates[ABusState]);
	}
	else
	{
		uint8_t s;
		for (s=0; s<2; ++s) if (strcmp(tok, ABusStates[s]) == 0) break;
		
		switch (s)
		{
			case ABUS_to_ZOUT:
				ABusState = ABUS_to_ZOUT;
				ABUS.Select(ABusState);
				Serial.println(MSG_OK);
				break;
				
			case ABUS_to_BIAS:
				ABusState = ABUS_to_BIAS;
				ABUS.Select(ABusState);
				Serial.println(MSG_OK);
				break;
				
			default:
				Serial.println(MSG_ERROR);
				break;
		}
	}
}

void handleACom()
{
	char *tok;
	
	tok = strtok(NULL, MSG_DELIMITER);
	
	if (tok == NULL)
	{
		Serial.println(AComStates[AComState]);
	}
	else
	{
		uint8_t s;
		for (s=0; s<2; ++s) if (strcmp(tok, AComStates[s]) == 0) break;
		
		switch (s)
		{
			case ACOM_to_ZIN:
				AComState = ACOM_to_ZIN;
				ACOM.Select(AComState);
				Serial.println(MSG_OK);
				break;
				
			case ACOM_to_GND:
				AComState = ACOM_to_GND;
				ACOM.Select(AComState);
				Serial.println(MSG_OK);
				break;
				
			default:
				Serial.println(MSG_ERROR);
				break;
		}
	}
}

void handleBias()
{
	char *tok;
	
	tok = strtok(NULL, MSG_DELIMITER);
	
	if (tok == NULL)
	{
		Serial.println(BiasStates[BiasState]);
	}
	else
	{
		uint8_t s;
		for (s=0; s<2; ++s) if (strcmp(tok, BiasStates[s]) == 0) break;
		
		switch (s)
		{
			case BIAS_to_INT:
				BiasState = BIAS_to_INT;
				BIAS.Select(BiasState);
				Serial.println(MSG_OK);
				break;
				
			case BIAS_to_EXT:
				BiasState = BIAS_to_EXT;
				BIAS.Select(BiasState);
				Serial.println(MSG_OK);
				break;
				
			default:
				Serial.println(MSG_ERROR);
				break;
		}
	}
}

void handleOnOff(bool value)
{
	char *tok;
	char buffer[5];
	bool newState[TRAP_N];
	
	tok = strtok(NULL, MSG_DELIMITER);
	
	if (tok == NULL)
	{
		for (uint8_t i=0; i<TRAP_N; ++i)
		{
			if (TrapState[i] == value)
			{
				getTrapName(buffer, i);
				Serial.print(buffer);
				Serial.print(" ");
			}
		}
		Serial.println(MSG_OK);
	}
	else
	{
		for (uint8_t i=0; i<TRAP_N; ++i)
		{
			newState[i] = TrapState[i];
		}
		
		int8_t trap;
		
		if (strcmp(tok, TRAP_CMD_ALL) == 0)
		{
			for (uint8_t i=0; i<TRAP_N; ++i)
			{
				TrapState[i] = value;
			}
			
			updateSwitches();
			
			Serial.println(MSG_OK);
			
			return;
		}
		
		while (tok != NULL)
		{
			trap = getTrap(tok);
			if (trap == -1)
			{
				Serial.println(MSG_ERROR);
				return;
			}
			
			newState[trap] = value;
			
			tok = strtok(NULL, MSG_DELIMITER);
		}
		
		for (uint8_t i=0; i<TRAP_N; ++i)
		{
			TrapState[i] = newState[i];
		}
		
		updateSwitches();
		
		Serial.println(MSG_OK);
	}
}

void handleOnly(bool value)
{
	char *tok;
	char buffer[5];
	bool newState[TRAP_N];
	
	tok = strtok(NULL, MSG_DELIMITER);
	
	if (tok == NULL)
	{
		Serial.println(MSG_ERROR);
	}
	else
	{
		for (uint8_t i=0; i<TRAP_N; ++i)
		{
			newState[i] = !value;
		}
		
		int8_t trap;
		
		while (tok != NULL)
		{
			trap = getTrap(tok);
			if (trap == -1)
			{
				Serial.println(MSG_ERROR);
				return;
			}
			
			newState[trap] = value;
			
			tok = strtok(NULL, MSG_DELIMITER);
		}
		
		for (uint8_t i=0; i<TRAP_N; ++i)
		{
			TrapState[i] = newState[i];
			
		}
		
		updateSwitches();
		
		Serial.println(MSG_OK);
	}
}

void handleStatus()
{
	char buffer[5];
	for (uint8_t i=0; i<TRAP_N; ++i)
	{
		getTrapName(buffer, i);
		Serial.print(buffer);
		if (TrapState[i])
		{
			Serial.println(" on");
		}
		else
		{
			Serial.println(" off");
		}
	}
	
	Serial.print("abus ");
	Serial.println(ABusStates[ABusState]);
	Serial.print("acom ");
	Serial.println(AComStates[AComState]);
	Serial.print("bias ");
	Serial.println(BiasStates[BiasState]);
}

void handleMode()
{
	char *tok;
	
	tok = strtok(NULL, MSG_DELIMITER);
	
	if (tok == NULL)
	{
		Serial.println(MSG_ERROR);
	}
	else
	{
		uint8_t s;
		for (s=0; s<3; ++s) if (strcmp(tok, Modes[s]) == 0) break;
		
		switch (s)
		{
			case MODE_EXC:
				AComState = ACOM_to_GND;
				ACOM.Select(AComState);
				
				ABusState = ABUS_to_BIAS;
				ABUS.Select(ABusState);
				
				BiasState = BIAS_to_INT;
				BIAS.Select(BiasState);
				
				Serial.println(MSG_OK);
				break;
				
			case MODE_Z:
				BiasState = BIAS_to_INT;
				BIAS.Select(BiasState);
				
				ABusState = ABUS_to_ZOUT;
				ABUS.Select(ABusState);
				
				AComState = ACOM_to_ZIN;
				ACOM.Select(AComState);
				
				Serial.println(MSG_OK);
				break;
				
			case MODE_DEP:
				AComState = ACOM_to_GND;
				ACOM.Select(AComState);
				
				ABusState = ABUS_to_BIAS;
				ABUS.Select(ABusState);
				
				BiasState = BIAS_to_EXT;
				BIAS.Select(BiasState);
				
				Serial.println(MSG_OK);
				break;
			
			default:
				Serial.println(MSG_ERROR);
				break;
		}
	}
}

void handleShort()
{
	// Set mode z
	BiasState = BIAS_to_INT;
	BIAS.Select(BiasState);
	
	ABusState = ABUS_to_ZOUT;
	ABUS.Select(ABusState);
	
	AComState = ACOM_to_ZIN;
	ACOM.Select(AComState);
	
	// off all
	for (uint8_t i=0; i<TRAP_N; ++i)
	{
		TrapState[i] = false;
	}
	
	// on short
	TrapState[TRAP_N-2] = true;
	
	updateSwitches();
	
	Serial.println(MSG_OK);
}

void handleOpen()
{
	// Set mode z
	BiasState = BIAS_to_INT;
	BIAS.Select(BiasState);
	
	ABusState = ABUS_to_ZOUT;
	ABUS.Select(ABusState);
	
	AComState = ACOM_to_ZIN;
	ACOM.Select(AComState);
	
	// off all
	for (uint8_t i=0; i<TRAP_N; ++i)
	{
		TrapState[i] = false;
	}
	
	// on open
	TrapState[TRAP_N-1] = true;
	
	updateSwitches();
	
	Serial.println(MSG_OK);
}

int8_t getSwitchNum(uint8_t site)
{
	if (site < 20)
	{
		return 19 - site;
	}

	return site;
}

void updateSwitches()
{
	uint8_t i;
	
	for (i=0; i<TRAP_N; ++i)
	{
		SwitchChain.Set(getSwitchNum(i), TrapState[i]);
	}
	SwitchChain.PrintConfiguration();
	SwitchChain.SendConfiguration();
	SwitchChain.PrintConfiguration();
	
}
